/*
 *  HD44780 LCD in 4-bit mode
 *
 *  DBx latched by falling edge of E, setup time 195 nsec.
 *  RS (and hardwired RW) latched by E rising edge, setup time 140 nsec
 *  E minimum cycle time 1 usec, minumum pulse width 450 nsec
 */
#include "standard.h"

extern const char ident[];

/* row starts for 2x16, 2x20, 2x40, 4x20, maybe other, modules */

STATIC word rowpos[] = { 0x00, 0x40, 0x14, 0x54 };
STATIC word row, col, nrows, ncols, linekink;

/*
 *  worst case delay during initialization
 *  and for Clear Display and Return Home, is 5 msec.
 *
 *  execution time _after_E_falls_
 *  is 120 usec worst case (with 160kHz osc ?)
 *  for commands other than Clear Display and Return Home.
 */
#define lcd_5msec() delay_ms(5)


STATIC byte xlat[0x100];

void
lcd_chr(byte c)
{
	if (++col == linekink)
		lcd_cmd(0xC0);

	lcd_data(xlat[c]);
}

void
lcd_chr_pretty(byte c)
{
	lcd_chr((c < ' ') ? 0xA5 : c); /* dot out unprintables */
}

void
lcd_str(const byte *s)
{
	while (*s)
		lcd_chr(*s++);
}

void
lcd_buf(const byte *s, word cc)
{
	while (cc--)
		lcd_chr(*s++);
}

void
lcd_buf_pretty(const byte *s, word cc)
{
	while (cc--)
		lcd_chr_pretty(*s++);
}

word
lcd_row(word which)
{
	if (which > nrows)
		which = nrows;

	row = which - 1;
	col = 0;
	lcd_cmd(0x80 | rowpos[row]); /* XXX check if needed at all */

	return (ncols); /* this much space to write into */
}

word
lcd_row_space()
{
	if (col < ncols)
		return (ncols - col);

	return (0);
}

void
lcd_finis()
{
	while (col < ncols)
		lcd_chr(' ');
}

STATIC const byte
	cg_check  [8] = { 0x01, 0x0A, 0x04, 0x00, 0x00, 0x00,  0x00, 0x00 },
	cg_padlock[8] = { 0x00, 0x00, 0x00, 0x00, 0x06, 0x09,  0x0F, 0x0F },
	cg_flag   [8] = { 0x00, 0x08, 0x0C, 0x0E, 0x0F, 0x0B,  0x09, 0x08 },
	cg_ake    [8] = { 0x04, 0x00, 0x0E, 0x01, 0x0F, 0x11,  0x0F, 0x00 },
	cg_zero   [8];

static void
chargen(word ix, const byte *bits)
{
	lcd_cmd(0x40 + ix*8);      /* CG RAM address */

	for (ix = 0; ix < 8; ++ix)
		lcd_data(*bits++);
}

/* RSSI - CG RAM - custom [6] */

STATIC word prevssi;

void
lcd_rssi(word rssi)
{
	rssi /= 50;       /* reasonable scale to [ 0 ... 8 ] */
	if (rssi > 8)
		rssi = 8;

	if (prevssi == rssi)
		return;

	if (prevssi < rssi) {         /* pile grows or shrinks ? */

		lcd_cmd(0x40 + 6*8 + 8 - rssi);
		do {
			lcd_data(0xFF);
		} while (++prevssi < rssi);
	} else {
		lcd_cmd(0x40 + 6*8 + 8 - prevssi);
		do {
			lcd_data(0x00);
		} while (--prevssi > rssi);
	}
}

static void
char_gen_preset()
{
	word i;

	chargen(3, cg_check);
	chargen(4, cg_padlock);
	chargen(5, cg_flag);
	chargen(6, cg_zero);  prevssi = 0;
	chargen(7, cg_ake);

	/* fill with unmapped characters */

	for (i = 0; i < sizeof(xlat); ++i)
		xlat[i] = i;

	/* latin-1 */

	xlat[0xC5] =
	xlat[0xE5] = 7;           /* Ake, custom in CG RAM [7] */

	xlat[0xC4] =
	xlat[0xE4] = 0x80 | 'a';  /* Aiti */

	xlat[0xD6] =
	xlat[0xF6] = 0x80 | 'o';  /* Oljy */

	xlat[0xDC] =
	xlat[0xFC] = 0x80 | 'u';  /* Umlaut */

	xlat[0xB0] = 0x80 | '_';  /* degree symbol */

	xlat['\\'] = 0x80 | '$';  /* backslash lookalike */
	xlat['~' ] = 0x80 | 'k';  /* tilde lookalike */

	/* DOS ? */

	xlat[0x81] =
	xlat[0x86] =
	xlat[0x8F] =
	xlat[0x8C] = 7;           /* Ake, custom in CG RAM [7] */

	xlat[0x80] =
	xlat[0x8A] =
	xlat[0x8E] =
	xlat[0x84] = 0x80 | 'a';  /* Aiti */

	xlat[0x85] =
	xlat[0x9A] =
	xlat[0x99] =
	xlat[0x94] = 0x80 | 'o';  /* Oljy */
}

void
lcd_init(byte *type)
{
	nrows = atoi(type + 0); /* type must be %1dx%d */
	ncols = atoi(type + 2);
	row = col = 0;

	linekink = (nrows == 1 && ncols == 16) ? 9 : 0;

	/*
	 *  Initialize LCD for 4bit I/F
	 *  It is assumed that 15 msec has passed since powerup.
	 */
	lcd_cmd8(0x38);  /* Function Set, in 8bit I/F, a couple of times. */
	lcd_5msec();
	lcd_cmd8(0x38);  /* Function Set, in 8bit I/F, a couple of times. */
	lcd_5msec();
	lcd_cmd8(0x38);  /* Function Set, in 8bit I/F, a couple of times. */
	lcd_5msec();

	lcd_cmd8(0x28);  /* Function Set, now entering 4bit I/F */
	lcd_5msec();

	/* 1x16 is not really single line XXX correct in all cases ? */

	lcd_cmd(0x20     /* Function Set DL=0 N= F=0 */
		| ((nrows != 1 || ncols == 16) ? 8 : 0));

	lcd_cmd(0x08);   /* Display OFF */
	lcd_cmd(0x0C);   /* Display ON  */
	lcd_cmd(0x06);   /* Entry Mode  I/D=1 S=0 increment, no shifting */
	lcd_cmd(0x02);   /* HOME */
	lcd_5msec();

	lcd_cmd(0x0F);   /* Display ON, Blink, Cursor on */
	lcd_cmd(0x0C);   /* Display ON, Blink OFF, Cursor OFF */

	char_gen_preset();

	lcd_row(4);
	lcd_str(ident + 4 + 6);
	lcd_finis();
}

